#version 420 core

uniform float fGlobalTime; // in seconds
uniform vec2 v2Resolution; // viewport resolution (in pixels)
uniform float fFrameTime; // duration of the last frame, in seconds

uniform sampler1D texFFT; // towards 0.0 is bass / lower freq, towards 1.0 is higher / treble freq
uniform sampler1D texFFTSmoothed; // this one has longer falloff and less harsh transients
uniform sampler1D texFFTIntegrated; // this is continually increasing
uniform sampler2D texPreviousFrame; // screenshot of the previous frame
uniform sampler2D texAcorn1;
uniform sampler2D texAcorn2;
uniform sampler2D texChecker;
uniform sampler2D texLeafs;
uniform sampler2D texNoise;
uniform sampler2D texRevisionBW;
uniform sampler2D texTex1;
uniform sampler2D texTex2;
uniform sampler2D texTex3;
uniform sampler2D texTex4;

// hi DiFFtY !!!!! <3 <3 <3

layout(r32ui) uniform coherent uimage2D[3] computeTex;
layout(r32ui) uniform coherent uimage2D[3] computeTexBack;

layout(location = 0) out vec4 out_color; // out_color must be written in order to see anything


float time=mod(fGlobalTime, 300);
vec2 res=v2Resolution;

vec3 rnd(vec3 s) {
  return fract(sin(s*453.234+s.yzx*546.389+s.zxy*985.345)*654.823);
}

void add(vec2 uv) {
  uv*=res.y;
  uv+=res*0.5;
  imageAtomicAdd(computeTex[0], ivec2(uv), 1);
}

float read(vec2 uv) {
  return imageLoad(computeTexBack[0], ivec2(uv)).x;
}

mat2 rot(float a) {
  float ca=cos(a);
  float sa=sin(a);
  return mat2(ca,sa,-sa,ca);
}

float map(vec3 p) {
  
  for(int i=0; i<3; ++i) {
    float t=time*0.01+texture(texFFTIntegrated, 0.02).x*0.2;
    p.xz*=rot(t);
    p.yz*=rot(t*0.8);
    p.xy=abs(p.xy)-1;
  }
  
  float d=length(p)-1.4;
  
  return abs(d);
}

vec2 proj(vec3 p) {
  
  float t=time*0.3;
  p.zy*=rot(t);
  p.zx*=rot(t*0.8);
  
  p.x += sin(time*0.2)*4;
  p.z-=10 + sin(time*0.3)*8;
  return p.xy/p.z;
}

void line(vec3 pa, vec3 pb) {
  vec2 a=proj(pa);
  vec2 b=proj(pb);
  vec2 d=abs(a-b)*res;
  float l=min(1000, max(d.x,d.y));
  for(int i=0; i<l; ++i) {
    add(mix(a,b,i/l));    
  }
}

vec3 axes(vec3 p1, vec3 dir) {
  vec3 p=p1+dir;
  vec2 off=vec2(0.01,0);
  float d=map(p);
  vec3 n=normalize(d-vec3(map(p-off.xyy), map(p-off.yxy), map(p-off.yyx)));
  p-=n*d;
  if(map(p)<0.1) {
    line(p1,p);
  }
  return p;
}

void main(void)
{
	vec2 uv = vec2(gl_FragCoord.x / v2Resolution.x, gl_FragCoord.y / v2Resolution.y);
	uv -= 0.5;
	uv /= vec2(v2Resolution.y / v2Resolution.x, 1);

  
  time = texture(texFFTIntegrated, 0.01).x*0.2;
  time+= rnd(vec3(floor(time/6-length(uv)*3))).x*400;
  
  vec3 col=vec3(0);
  
  vec2 fuv=gl_FragCoord.xy;
  
  float num=3+rnd(vec3(uv, floor(time/10))).x*10;
  if(fuv.y<num) {
    vec3 p=rnd(vec3(uv, 1))-0.5;
    p*=10;
    if(map(p)<0.1) {
      vec3 h=rnd(vec3(uv, 2))-0.5;
      vec3 l=p;
      vec3 n=p;
      for(int i=0; i<5; ++i) {
        vec3 h=rnd(vec3(uv, 2+i))-0.5;
        n+=h*(0.02 + rnd(floor((l*0.01+time*0.3)*0.1)).x*3);
        float g=0.01 + rnd(floor((l*0.3+time*0.06)*0.3)).x*3;
        n.xy*=rot(time*0.1*g);
        n.zy*=rot(time*0.2*g);
        l=axes(l,n);
      }
    }
  }
  
  
 // col += texture(texFFTSmoothed, abs(uv.x)*0.1).x*10;
  
  col += read(fuv) * vec3(1,0.5,0.3 + abs(uv.y)*3) * 0.3;
  col += texture(texPreviousFrame, fuv/res).xyz*vec3(0.9,0.5,0.3);
  
	out_color = vec4(col, 1);
}